
/*  XTension Smart Meter Reader Firmare
    Version 1.9 originally by James Sentman james@sentman.com
    www.MacHomeAutomation.com
    
    This particular instance of the software is written by James Sentman
    but contains many tips and techniques from other published projects that are 
    similar and so everything you learn here or can take away from here is
    offered as open source and you may do with it as you please.


*/

#include <EEPROM.h>


//005 begins sending only every 5 seconds instead of all the time.
//006 adds filtering for ludicrious pulses.

//individual pulses beyond this will be ignored.
int outOfBandLimit = 18000; //150 amps? 
unsigned int outOfBandCount = 0; //sends a count when things get back to normal

//analog input from phototransistor
int sensorPin = A0;

//internal LED pin
int ledPin = 13;

//for storing sensor value
int sensorValue = 0;

//initial threashold values for sensing on and offs
//these are reset upon powerup by the scan routine
int lowThreshold = 80;
int highThreshold = 200;

//is the pulse currently on or off
boolean currentState = false;

//this allows us to skip the first pulse as the 
//timing values wont be initted yet and will have garbage
//until one is received
boolean gotFirstPulse = false;

//Number of pulses, used to measure energy.
long pulseCount = 0;   

//Used to measure power.
unsigned long pulseTime,lastTime,intervalTime, pulseWidth, pulseStart, lastBufferAdd;

//power and energy
double power, elapsedkWh;

//Number of pulses per wh
//in every meter I've heard of this is one pulse per watt hour, but yours might be different
int ppwh = 1; //1000 pulses/kwh = 1 pulse per wh


volatile boolean GotPulse;


//this array is used to sample the values and 
//come up with a middle point moving average
//that can be used to  make some decisions on on/off point

int movingAverage[350];
int currentMidpoint = 0;
int currentMin = 1000;
int currentMax = 1;

int rollingIndex = 0;


//for when data is sent to XTension
int pulsesThisFrame = 0;
double pulseFrameAccumulator = 0;
double pulseFrameAverage = 0;
double pulseFrameUsedAccumulator = 0;
double pulseFramePeak = 0;
unsigned long pulseFrameLastTransmit = 0;




void setup(){
  pinMode( ledPin, OUTPUT);
  Serial.begin( 57600);
  
  //add a delay for the xBee to associated on the network
  digitalWrite( ledPin, HIGH);
  delay(3000);
  digitalWrite( ledPin, LOW);
  
  
  Serial.println( "XTension Smart Meter Reader version 2.0 Startup.");

  buildMovingAverage();
  makeAverage();
  lastBufferAdd = millis();
  pulseFrameLastTransmit = millis(); //wait 5 seconds to send.
}

void loop(){
  sensorValue = analogRead( sensorPin);
   
  if (sensorValue < lowThreshold){
    if (! currentState){
      currentState = true;
      digitalWrite( ledPin, HIGH);
     
 
 //
 // A pulse has been received
 // calculate and send values
 
      doPulse();
      
    } //change of state
  } else {
   if (sensorValue > highThreshold){
     if (currentState){
       currentState = false;
       digitalWrite( ledPin, LOW);
      }  
   } 
  }
  
  //manage the rolling average
  if (millis() - lastBufferAdd > 15){
    lastBufferAdd = millis();
    rollingIndex++;
    if (rollingIndex > 349){
      rollingIndex = 0;
    }
  
    movingAverage[ rollingIndex] = sensorValue;
  
    makeAverage();

  }
  
  //check for a command from XTension
  //single character commands to keep from
  //bigging down the CPU and causing timing errors
  //I = a ping, it replies with "HELLO"
  //R = resets the counters, send every midnight
  //S = do a scan and just send the values to the log
  //  good for checking for goofy data or noise.
  
    if (Serial.available()) {
    // read the most recent byte
      int thisChar = Serial.read();
      switch (thisChar){
        case 73: //I is ping
          //a ping request, just sent it back
          Serial.println( "PING");
          break;
        
        case 82: //R to reset
          //pulseCount = 0;
          //pulseTime = micros();
          //lastTime = pulseTime;
          //gotFirstPulse = false;
          elapsedkWh = 0;
          pulseCount = 0;
          Serial.println( "RESET");
          break;
                  
        case 83: //S scan values
          Serial.println( "RESCAN");
          doScan();
          break;
      }
    }
}

void doPulse(){
  
  //check for faux out of band pulses
  double TestPower;
  TestPower = (3600000000.0 / (micros() - pulseTime))/ppwh;
  
  if (TestPower > 22000.0){
    
    outOfBandCount++;

    return;
  }
    
  
   lastTime = pulseTime;
   pulseTime = micros();
  
  if (lastTime > pulseTime){
    //micros rollover
    //Serial.println( "micros rollover");
  }

  //pulseCounter
  pulseCount++;

  //Calculate power
  intervalTime = pulseTime - lastTime;
  power = (3600000000.0 / intervalTime) / ppwh;
  
  //Find kwh elapsed
  elapsedkWh = (1.0 * pulseCount / (ppwh * 1000)); //multiply by 1000 to pulses per wh to kwh convert wh to kwh
  
  //update the 5 second buffer accumulators
  pulsesThisFrame++;
  pulseFrameAccumulator = pulseFrameAccumulator + power;
  pulseFramePeak = max( pulseFramePeak, power);
  
  if (outOfBandCount > 0){
    Serial.print( "out of band pulses ");
    Serial.println( outOfBandCount);
    outOfBandCount = 0;
  }
  

//if it's the first pulse then dont send this as it will be bogus, wait for the second
  if (gotFirstPulse) {
    
    //send info only every 5 seconds
    //subtract time and check milliseconds.
    if ((pulseTime - pulseFrameLastTransmit) > 5000000){
//    
//  it's been more than 5 seconds since it last transmitted
//  so now we average the data and send it
//
      pulseFrameAverage = pulseFrameAccumulator / pulsesThisFrame;
    
//
//  send P then the average power usage the last 5 seconds then the elapsed count of kwh and finally the peak output
    
    Serial.print( "instant=");
      Serial.println( pulseFrameAverage, DEC);    
      
      Serial.print( "whtotal=");
      Serial.println( pulseCount, DEC);
      
      Serial.print( "peak=");
      Serial.println( pulseFramePeak, DEC);
      
      Serial.print( "whinstant=");
      Serial.println( pulsesThisFrame, DEC);
      
    
 //
 //  now reset pulseAccumulators
 //
 
     pulseFrameAccumulator = 0;
     pulsesThisFrame = 0;
     pulseFrameLastTransmit = pulseTime;
     pulseFramePeak = 0;
   }  

  } else {
    gotFirstPulse = true;
    Serial.println( "FIRSTPULSE");
  }

}

//called upon startup to build the moving average index and
//begin the sampling

void buildMovingAverage(){

  Serial.println( "scan for moving average value...");

  
  for (int i=0; i<350; i++){
    movingAverage[i] = analogRead( A0);
    delay( 25);
  }
 

  
  
}

void makeAverage(){
 int collector = 0;
 int tempMidpoint = 0;
 int workInt = 0;
 
 currentMin = 1000;
 currentMax = 1;
  
 for (int i=0; i<350; i++){
  // collector = collector + movingAverage[i];
   currentMin = min( currentMin, movingAverage[i]);
   currentMax = max( currentMax, movingAverage[i]);
 }
 
 tempMidpoint = (currentMax - currentMin) / 4; //quarter value of different is cutoff
 

 workInt = (currentMin + tempMidpoint);
 
 if (lowThreshold != workInt){
   lowThreshold = workInt;
  // Serial.print( "M low threshold set: " );
  // Serial.println( lowThreshold);
 }
 
 
// workInt = (currentMin + (tempMidpoint + 25));
workInt = (currentMax - tempMidpoint);
 
 
 if (highThreshold != workInt){
   highThreshold = workInt;
 //  Serial.print( "M high threshold set: " );
 //  Serial.println( highThreshold);
 }
 
 
  
}
  
  void doScan(){
 
 //called from the S command and outputs a few seconds of readings and then min/max them.

int minValue = 10000;
int maxValue = -10000;
int workValue = 0;

  for (int i=0; i < 500; i++){
    workValue = analogRead( A0);
    minValue = min( minValue, workValue);
    maxValue = max( maxValue, workValue);
    
//    Serial.print( "S ");
    Serial.println( workValue);
    delay(100);
  }
  
  Serial.print( "max ");
  Serial.print( maxValue);
  Serial.print( " min ");
  Serial.println( minValue);
  

}


/*-----------------------------------------------------------------*/
//  E E P R O M  

void eepromWriteDouble( int address, const double newValue){
	const byte* p = (const byte*)(const void*)&newValue;
	int i;
	for (i=0; i<4; i++){
		EEPROM.write( address++, *p++);
	}

}


/*-----------------------------------------------------------------*/
//  E E P R O M   R E A D   D O U B L E



double eepromReadDouble( int address){
double oldValue = 0.0;

  byte* p = (byte*)(void*)&oldValue;
  int i;
  for (i=0; i<4; i++){
    *p++ = EEPROM.read( address++);
  }
  
  return oldValue;
}

